from copy import deepcopy
from dataclasses import dataclass, field
import datetime
import inspect
import typing as t

from piccolo.columns import Column, OnDelete, OnUpdate, column_types

from piccolo.custom_types import DatetimeDefault
from piccolo.engine import engine_finder
from piccolo.migrations.auto.diffable_table import DiffableTable
from piccolo.migrations.auto.operations import (
    DropColumn,
    RenameColumn,
    AlterColumn,
    RenameTable,
)
from piccolo.table import Table


@dataclass
class AddColumnClass:
    column: Column
    table_class_name: str
    tablename: str


@dataclass
class AddColumnCollection:
    add_columns: t.List[AddColumnClass] = field(default_factory=list)

    def append(self, add_column: AddColumnClass):
        self.add_columns.append(add_column)

    def for_table_class_name(
        self, table_class_name: str
    ) -> t.List[AddColumnClass]:
        return [
            i
            for i in self.add_columns
            if i.table_class_name == table_class_name
        ]

    def columns_for_table_class_name(
        self, table_class_name: str
    ) -> t.List[Column]:
        return [
            i.column
            for i in self.add_columns
            if i.table_class_name == table_class_name
        ]

    @property
    def table_class_names(self) -> t.List[str]:
        return list(set([i.table_class_name for i in self.add_columns]))


@dataclass
class DropColumnCollection:
    drop_columns: t.List[DropColumn] = field(default_factory=list)

    def append(self, drop_column: DropColumn):
        self.drop_columns.append(drop_column)

    def for_table_class_name(self, table_class_name: str) -> t.List[str]:
        return [
            i.column_name
            for i in self.drop_columns
            if i.table_class_name == table_class_name
        ]

    @property
    def table_class_names(self) -> t.List[str]:
        return list(set([i.table_class_name for i in self.drop_columns]))


@dataclass
class RenameColumnCollection:
    rename_columns: t.List[RenameColumn] = field(default_factory=list)

    def append(self, rename_column: RenameColumn):
        self.rename_columns.append(rename_column)

    def for_table_class_name(
        self, table_class_name: str
    ) -> t.List[RenameColumn]:
        return [
            i
            for i in self.rename_columns
            if i.table_class_name == table_class_name
        ]

    @property
    def table_class_names(self) -> t.List[str]:
        return list(set([i.table_class_name for i in self.rename_columns]))


@dataclass
class AlterColumnCollection:
    alter_columns: t.List[AlterColumn] = field(default_factory=list)

    def append(self, alter_column: AlterColumn):
        self.alter_columns.append(alter_column)

    def for_table_class_name(
        self, table_class_name: str
    ) -> t.List[AlterColumn]:
        return [
            i
            for i in self.alter_columns
            if i.table_class_name == table_class_name
        ]

    @property
    def table_class_names(self) -> t.List[str]:
        return list(set([i.column_name for i in self.alter_columns]))


@dataclass
class MigrationManager:
    """
    Each auto generated migration returns a MigrationManager. It contains
    all of the schema changes that migration wants to make.
    """

    migration_id: str = ""
    add_tables: t.List[DiffableTable] = field(default_factory=list)
    drop_tables: t.List[DiffableTable] = field(default_factory=list)
    rename_tables: t.List[RenameTable] = field(default_factory=list)
    add_columns: AddColumnCollection = field(
        default_factory=AddColumnCollection
    )
    drop_columns: DropColumnCollection = field(
        default_factory=DropColumnCollection
    )
    rename_columns: RenameColumnCollection = field(
        default_factory=RenameColumnCollection
    )
    alter_columns: AlterColumnCollection = field(
        default_factory=AlterColumnCollection
    )
    raw: t.List[t.Union[t.Callable, t.Coroutine]] = field(default_factory=list)

    def add_table(
        self,
        class_name: str,
        tablename: str,
        columns: t.Optional[t.List[Column]] = None,
    ):
        if not columns:
            columns = []

        self.add_tables.append(
            DiffableTable(
                class_name=class_name, tablename=tablename, columns=columns
            )
        )

    def drop_table(self, class_name: str, tablename: str):
        self.drop_tables.append(
            DiffableTable(class_name=class_name, tablename=tablename)
        )

    def rename_table(
        self, old_class_name: str, new_class_name: str, new_tablename: str
    ):
        self.rename_tables.append(
            RenameTable(
                old_class_name=old_class_name,
                new_class_name=new_class_name,
                new_tablename=new_tablename,
            )
        )

    def add_column(
        self,
        table_class_name: str,
        tablename: str,
        column_name: str,
        column_class_name: str,
        params: t.Dict[str, t.Any] = {},
    ):
        column_class = getattr(column_types, column_class_name)
        cleaned_params = self.deserialise_params(params)
        column = column_class(**cleaned_params)
        column._meta.name = column_name
        self.add_columns.append(
            AddColumnClass(
                column=column,
                tablename=tablename,
                table_class_name=table_class_name,
            )
        )

    def drop_column(
        self, table_class_name: str, tablename: str, column_name: str
    ):
        self.drop_columns.append(
            DropColumn(
                table_class_name=table_class_name,
                column_name=column_name,
                tablename=tablename,
            )
        )

    def rename_column(
        self,
        table_class_name: str,
        tablename: str,
        old_column_name: str,
        new_column_name: str,
    ):
        self.rename_columns.append(
            RenameColumn(
                table_class_name=table_class_name,
                tablename=tablename,
                old_column_name=old_column_name,
                new_column_name=new_column_name,
            )
        )

    def alter_column(
        self,
        table_class_name: str,
        tablename: str,
        column_name: str,
        params: t.Dict[str, t.Any],
    ):
        """
        All possible alterations aren't currently supported.
        """
        self.alter_columns.append(
            AlterColumn(
                table_class_name=table_class_name,
                tablename=tablename,
                column_name=column_name,
                params=params,
            )
        )

    def add_raw(self, raw: t.Union[t.Callable, t.Coroutine]):
        """
        A migration manager can execute arbitrary functions or coroutines when
        run. This is useful if you want to execute raw SQL.
        """
        self.raw.append(raw)

    ###########################################################################

    def deserialise_params(
        self, params: t.Dict[str, t.Any]
    ) -> t.Dict[str, t.Any]:
        """
        When reading column params from a migration file, we need to convert
        them from their serialised form.
        """
        params = deepcopy(params)

        references = params.get("references")
        if references:
            components = references.split("|")
            if len(components) == 1:
                class_name = components[0]
                tablename = None
            elif len(components) == 2:
                class_name, tablename = components
            else:
                raise ValueError(
                    "Unrecognised Table serialisation - should either be "
                    "`SomeClassName` or `SomeClassName|some_table_name`."
                )

            _Table: t.Type[Table] = type(
                references, (Table,), {},
            )
            if tablename:
                _Table._meta.tablename = tablename
            params["references"] = _Table

        on_delete = params.get("on_delete")
        if on_delete:
            enum_name, item_name = on_delete.split(".")
            if enum_name == "OnDelete":
                params["on_delete"] = getattr(OnDelete, item_name)

        on_update = params.get("on_update")
        if on_update:
            enum_name, item_name = on_update.split(".")
            if enum_name == "OnUpdate":
                params["on_update"] = getattr(OnUpdate, item_name)

        default = params.get("default")
        if isinstance(default, str):
            if default.startswith("DatetimeDefault"):
                _, item_name = default.split(".")
                params["default"] = getattr(DatetimeDefault, item_name)
            else:
                try:
                    params["default"] = datetime.datetime.fromisoformat(
                        default
                    )
                except ValueError:
                    pass

        return params

    ###########################################################################

    async def run(self):
        print("Running MigrationManager ...")

        engine = engine_finder()

        if not engine:
            raise Exception("Can't find engine")

        async with engine.transaction():

            for raw in self.raw:
                if inspect.iscoroutinefunction(raw):
                    await raw()
                else:
                    raw()

            ###################################################################
            # Add tables

            for table in self.add_tables:
                columns = self.add_columns.columns_for_table_class_name(
                    table.class_name
                )
                _Table: t.Type[Table] = type(
                    table.class_name,
                    (Table,),
                    {column._meta.name: column for column in columns},
                )
                _Table._meta.tablename = table.tablename

                await _Table.create_table().run()

            ###################################################################
            # Drop tables

            for table in self.drop_tables:
                await table.to_table_class().alter().drop_table().run()

            ###################################################################
            # Rename tables

            for rename_table in self.rename_tables:
                _Table: t.Type[Table] = type(
                    rename_table.old_class_name, (Table,), {}
                )
                await _Table.alter().rename_table(
                    new_name=rename_table.new_tablename
                ).run()

            ###################################################################
            # Add columns, which belong to existing tables

            new_table_class_names = [i.class_name for i in self.add_tables]

            for table_class_name in self.add_columns.table_class_names:
                if table_class_name in new_table_class_names:
                    continue

                add_columns: t.List[
                    AddColumnClass
                ] = self.add_columns.for_table_class_name(table_class_name)

                _Table: t.Type[Table] = type(
                    add_columns[0].table_class_name, (Table,), {}
                )
                _Table._meta.tablename = add_columns[0].tablename

                for add_column in add_columns:
                    column = add_column.column
                    await _Table.alter().add_column(
                        name=column._meta.name, column=column
                    ).run()

            ###################################################################
            # Drop columns

            for table_class_name in self.drop_columns.table_class_names:
                columns = self.drop_columns.for_table_class_name(
                    table_class_name
                )

                if not columns:
                    continue

                _Table: t.Type[Table] = type(table_class_name, (Table,), {})
                _Table._meta.tablename = columns[0].tablename

                for column in columns:
                    await _Table.alter().drop_column(
                        column=column.column_name
                    ).run()

            ###################################################################
            # Rename columns

            for table_class_name in self.rename_columns.table_class_names:
                columns = self.rename_columns.for_table_class_name(
                    table_class_name
                )

                if not columns:
                    continue

                _Table: t.Type[Table] = type(table_class_name, (Table,), {})
                _Table._meta.tablename = columns[0].tablename

                for column in columns:
                    await _Table.alter().rename_column(
                        column=column.old_column_name,
                        new_name=column.new_column_name,
                    ).run()

            ###################################################################
            # Alter columns

            for table_class_name in self.alter_columns.table_class_names:
                alter_columns = self.alter_columns.for_table_class_name(
                    table_class_name
                )

                if not alter_columns:
                    continue

                _Table: t.Type[Table] = type(table_class_name, (Table,), {})
                _Table._meta.tablename = alter_columns[0].tablename

                for column in alter_columns:
                    params = column.params
                    row_name = column.row_name

                    null = params.get("null")
                    if null is not None:
                        await _Table.alter().set_null(
                            column=row_name, boolean=null
                        ).run()

                    length = params.get("length")
                    if length is not None:
                        await _Table.alter().set_length(
                            column=row_name, length=length
                        ).run()

                    unique = params.get("unique")
                    if unique is not None:
                        await _Table.alter().set_unique(
                            column=row_name, boolean=unique
                        ).run()



from __future__ import annotations
from dataclasses import dataclass, field
from importlib import import_module
import typing as t

if t.TYPE_CHECKING:
    from piccolo.table import Table


@dataclass
class AppConfig:
    app_name: str
    migrations_folder_path: str
    table_classes: t.List[t.Type[Table]] = field(default_factory=list)
    migration_dependencies: t.List[str] = field(default_factory=list)

    def register_table(self, table_class: t.Type[Table]):
        self.table_classes.append(table_class)
        return table_class


@dataclass
class AppRegistry:
    apps: t.List[str] = field(default_factory=list)

    def __post_init__(self):
        self.app_configs: t.Dict[str, AppConfig] = {}
        for app in self.apps:
            app_conf_module = import_module(app)
            app_config: AppConfig = getattr(app_conf_module, "APP_CONFIG")
            self.app_configs[app_config.app_name] = app_config

    def get_app_config(self, app_name: str) -> t.Optional[AppConfig]:
        return self.app_configs.get(app_name)

    def get_table_classes(self, app_name: str):
        """
        Returns each Table subclass defined in the given app.
        """
        app_config = self.get_app_config(app_name=app_name)
        if not app_config:
            raise ValueError(f"Unrecognised app_name: {app_name}")
        return app_config.table_classes



from piccolo import table
from piccolo import columns


###############################################################################
# Simple example


class Manager(table.Table):
    name = columns.Varchar(length=50)


class Band(table.Table):
    name = columns.Varchar(length=50)
    manager = columns.ForeignKey(Manager, null=True)
    popularity = columns.Integer(default=0)


###############################################################################
# More complex


class Venue(table.Table):
    name = columns.Varchar(length=100)
    capacity = columns.Integer(default=0)


class Concert(table.Table):
    band_1 = columns.ForeignKey(Band)
    band_2 = columns.ForeignKey(Band)
    venue = columns.ForeignKey(Venue)



from ..base import DBTestCase
from ..example_project.tables import Band


class TestUpdate(DBTestCase):
    def test_update(self):
        self.insert_rows()

        Band.update({Band.name: "Pythonistas3"}).where(
            Band.name == "Pythonistas"
        ).run_sync()

        response = (
            Band.select(Band.name)
            .where(Band.name == "Pythonistas3")
            .run_sync()
        )
        print(f"response = {response}")

        self.assertEqual(response, [{"name": "Pythonistas3"}])

    def test_update_values(self):
        self.insert_rows()

        Band.update({Band.name: "Pythonistas3"}).where(
            Band.name == "Pythonistas"
        ).run_sync()

        response = (
            Band.select(Band.name)
            .where(Band.name == "Pythonistas3")
            .run_sync()
        )
        print(f"response = {response}")

        self.assertEqual(response, [{"name": "Pythonistas3"}])


class TestUpdateOperators(DBTestCase):
    def test_add(self):
        self.insert_row()

        Band.update({Band.popularity: Band.popularity + 10}).run_sync()

        response = Band.select(Band.popularity).first().run_sync()

        self.assertEqual(response["popularity"], 1010)

    def test_radd(self):
        self.insert_row()

        Band.update({Band.popularity: 10 + Band.popularity}).run_sync()

        response = Band.select(Band.popularity).first().run_sync()

        self.assertEqual(response["popularity"], 1010)

    def test_sub(self):
        self.insert_row()

        Band.update({Band.popularity: Band.popularity - 10}).run_sync()

        response = Band.select(Band.popularity).first().run_sync()

        self.assertEqual(response["popularity"], 990)

    def test_rsub(self):
        self.insert_row()

        Band.update({Band.popularity: 1100 - Band.popularity}).run_sync()

        response = Band.select(Band.popularity).first().run_sync()

        self.assertEqual(response["popularity"], 100)

    def test_mul(self):
        self.insert_row()

        Band.update({Band.popularity: Band.popularity * 2}).run_sync()

        response = Band.select(Band.popularity).first().run_sync()

        self.assertEqual(response["popularity"], 2000)

    def test_rmul(self):
        self.insert_row()

        Band.update({Band.popularity: 2 * Band.popularity}).run_sync()

        response = Band.select(Band.popularity).first().run_sync()

        self.assertEqual(response["popularity"], 2000)

    def test_div(self):
        self.insert_row()

        Band.update({Band.popularity: Band.popularity / 10}).run_sync()

        response = Band.select(Band.popularity).first().run_sync()

        self.assertEqual(response["popularity"], 100)

    def test_rdiv(self):
        self.insert_row()

        Band.update({Band.popularity: 1000 / Band.popularity}).run_sync()

        response = Band.select(Band.popularity).first().run_sync()

        self.assertEqual(response["popularity"], 1)



class Operator:
    template = ""


class IsNull(Operator):
    template = "{name} IS NULL"


class Equal(Operator):
    template = "{name} = {value}"


class NotEqual(Operator):
    template = "{name} != {value}"


class In(Operator):
    template = "{name} IN ({values})"


class NotIn(Operator):
    template = "{name} NOT IN ({values})"


class Like(Operator):
    template = "{name} LIKE {value}"


class ILike(Operator):
    template = "{name} ILIKE {value}"


class NotLike(Operator):
    template = "{name} NOT LIKE {value}"


class GreaterThan(Operator):
    # Add permitted types???
    template = "{name} > {value}"


class GreaterEqualThan(Operator):
    template = "{name} >= {value}"


class LessThan(Operator):
    template = "{name} < {value}"


class LessEqualThan(Operator):
    template = "{name} <= {value}"



import jinja2


TEMPLATE = """
from piccolo.migrations.auto import MigrationManager


ID = '{{ migration_id }}'


async def forwards():
    manager = MigrationManager(migration_id=ID)
    {% if auto %}
    {% for alter_statement in alter_statements %}
    {{ alter_statement }}
    {% endfor %}
    {% else %}
    def run():
        print(f"running {ID}")

    manager.add_raw(run)
    {% endif %}
    return manager


async def backwards():
    pass

"""


def render_template(**kwargs):
    template = jinja2.Template(TEMPLATE, trim_blocks=True, lstrip_blocks=True)
    return template.render(**kwargs)



import asyncio

from piccolo.migrations.auto import MigrationManager

from tests.base import DBTestCase


class TestMigrationManager(DBTestCase):
    def test_rename_column(self):
        """
        Test running a MigrationManager which contains a column rename
        operation.
        """
        self.insert_row()

        manager = MigrationManager()
        manager.rename_column(
            table_class_name="Band",
            tablename="band",
            old_column_name="name",
            new_column_name="title",
        )
        asyncio.run(manager.run())

        response = self.run_sync("SELECT * FROM band;")
        self.assertTrue("title" in response[0].keys())
        self.assertTrue("name" not in response[0].keys())

    def test_raw_function(self):
        """
        Test adding raw functions to a MigrationManager.
        """

        class HasRun(Exception):
            pass

        def run():
            raise HasRun("I was run!")

        manager = MigrationManager()
        manager.add_raw(run)

        with self.assertRaises(HasRun):
            asyncio.run(manager.run())

    def test_raw_coroutine(self):
        """
        Test adding raw coroutines to a MigrationManager.
        """

        class HasRun(Exception):
            pass

        async def run():
            raise HasRun("I was run!")

        manager = MigrationManager()
        manager.add_raw(run)

        with self.assertRaises(HasRun):
            asyncio.run(manager.run())



from __future__ import annotations
import asyncio
import datetime
import os
import sys
import typing as t
from types import ModuleType

import black
import click

from piccolo.commands.migration.base import (
    BaseMigrationManager,
    MigrationModule,
)
from piccolo.conf.apps import AppConfig, AppRegistry
from piccolo.migrations.auto import (
    SchemaSnapshot,
    MigrationManager,
    DiffableTable,
    SchemaDiffer,
)
from piccolo.migrations.template import render_template


MIGRATION_MODULES: t.Dict[str, ModuleType] = {}


def _create_migrations_folder(migrations_path: str) -> bool:
    """
    Creates the folder that migrations live in. Returns True/False depending
    on whether it was created or not.
    """
    if os.path.exists(migrations_path):
        return False
    else:
        os.mkdir(migrations_path)
        for filename in ("__init__.py", "config.py"):
            with open(os.path.join(migrations_path, filename), "w"):
                pass
        return True


def _create_new_migration(app_config: AppConfig, auto=False) -> None:
    """
    Creates a new migration file on disk.
    """
    _id = datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
    path = os.path.join(app_config.migrations_folder_path, f"{_id}.py")

    if auto:
        alter_statements = AutoMigrationManager().get_alter_statements(
            app_config=app_config
        )

        if len(alter_statements) == 0:
            print("No changes detected - exiting.")
            sys.exit(1)

        file_contents = render_template(
            migration_id=_id, auto=True, alter_statements=alter_statements,
        )
    else:
        file_contents = render_template(migration_id=_id, auto=False)

    # Beautify the file contents a bit.
    file_contents = black.format_str(
        file_contents, mode=black.FileMode(line_length=82)
    )

    with open(path, "w") as f:
        f.write(file_contents)


###############################################################################


class AutoMigrationManager(BaseMigrationManager):
    def get_alter_statements(self, app_config: AppConfig) -> t.List[str]:
        """
        Works out which alter statements are required.
        """
        alter_statements: t.List[str] = []

        migration_managers = self.get_migration_managers(
            app_name=app_config.app_name
        )

        schema_snapshot = SchemaSnapshot(migration_managers)
        snapshot = schema_snapshot.get_snapshot()

        # Now get the current schema:
        current_diffable_tables = [
            DiffableTable(
                class_name=i.__name__,
                tablename=i._meta.tablename,
                columns=i._meta.non_default_columns,
            )
            for i in app_config.table_classes
        ]

        # Compare the current schema with the snapshot
        differ = SchemaDiffer(
            schema=current_diffable_tables, schema_snapshot=snapshot
        )
        alter_statements = differ.get_alter_statements()

        return alter_statements


###############################################################################


@click.argument("app_name")
@click.option(
    "--auto", is_flag=True, help="Auto create the migration contents."
)
@click.command()
def new(app_name: str, auto: bool):
    """
    Creates a new file like piccolo_migrations/2018-09-04T19:44:09.py
    """
    print("Creating new migration ...")

    try:
        import piccolo_conf
    except ImportError:
        print("Can't find piccolo_conf")
        sys.exit(1)

    try:
        app_registry: AppRegistry = piccolo_conf.APP_REGISTRY
    except AttributeError:
        print("APP_REGISTRY isn't defined in piccolo_conf")
        sys.exit(1)

    app_config = app_registry.get_app_config(app_name)

    if not app_config:
        raise ValueError(f"Unrecognised app_name: {app_name}")

    _create_migrations_folder(app_config.migrations_folder_path)
    _create_new_migration(app_config=app_config, auto=auto)
